package require tclpy
package require oo::util

oo::class create ::pyman {
    variable pythonObject

    constructor {args} {
        variable pythonObject

        set className [info object class [self]]
        if {$className eq "::pyman"} {
            throw [list PYMAN LOOKUP SUBCOMMAND] {unknown subcommand must be: import or addPath}
        }

        set tclModuleName [join [lrange [split $className {:}] 0 end-2] {:}]
        set pythonModuleName [set ${tclModuleName}::pythonModuleName]
        set pythonClassName [lindex [split $className {:}] end]
        set pythonObject [string map {{::} {_}} [self]]

        ##
        ## Define a function to create the object and assign it to a global variable
        ##
        set defineInstance {def pymanDefineInstance( objName}
        set createString {globals()[objName] = }
        append createString "$pythonModuleName.${pythonClassName}("
        set argCnt [llength $args]
        for {set i 1} {$i <= $argCnt} {incr i} {
            append defineInstance ", x$i"
            if {$i > 1} {
                append createString ", "
            }
            append createString x$i
        }
        append defineInstance "):"
        append defineInstance "\n    " $createString ")"
        append defineInstance "\n    " "return"

        py eval $defineInstance
        py call pymanDefineInstance $pythonObject {*}$args
    }

    classmethod import {args} {
        switch -exact -- [llength $args] {
            1   {
                ##
                ## Just a moduleName
                ##
                set moduleName [lindex $args 0]
                set namespaceName [format {::pyman::%s} $moduleName]
            }
            3 {
                ##
                ## We have -namespace namespaceName moudleName
                ##
                if {[lindex $args 0] ne "-namespace"} {
                    throw [list PYMAN BADIMP $args] "Bad usage must be: pyman import ?-namespace namespaceName? moduleName"
                }
                set namespaceName [lindex $args 1]
                set moduleName [lindex $args 2]
            }
            default {
                ##
                ## Illegal usage
                ##
                throw [list PYMAN BADIMP $args] "Bad usage must be: pyman import ?-namespace namespaceName? moduleName"
            }
        }

        ##
        ## Format for message bodies
        ##
        set methodBodyFmt {
            variable pythonObject

            py call $pythonObject.%1$s {*}$args
        }

        ##
        ## Have python import the module
        ##
        py import $moduleName

        ##
        ## Create the Tcl namespace that maps to the imported module
        ##
        namespace eval $namespaceName {}
        set ${namespaceName}::pythonModuleName $moduleName

        ##
        ## Create the inspection macro and look at the contents of the module
        ##
        set getModuleInfoFmt {def pymanInspect():}
        append getModuleInfoFmt "\n\t" {results = []}
        append getModuleInfoFmt "\n\t" {for name, data in inspect.getmembers(%1$s):}
        append getModuleInfoFmt "\n\t\t" {temp = []}
        append getModuleInfoFmt "\n\t\t" {temp.append(name)}
        append getModuleInfoFmt "\n\t\t" {temp.append(repr(data))}
        append getModuleInfoFmt "\n\t\t" {results.append(temp)}
        append getModuleInfoFmt "\n\t" {return results}

        set pymanGetAttributesFmt {def pymanGetAttributes():}
        append pymanGetAttributesFmt "\n\t" {results = []}
        append pymanGetAttributesFmt "\n\t" {for name, data in inspect.getmembers(%1$s.%2$s, lambda a:not(inspect.isroutine(a))):}
        append pymanGetAttributesFmt "\n\t\t" {temp = []}
        append pymanGetAttributesFmt "\n\t\t" {temp.append(name)}
        append pymanGetAttributesFmt "\n\t\t" {temp.append(repr(data))}
        append pymanGetAttributesFmt "\n\t\t" {results.append(temp)}
        append pymanGetAttributesFmt "\n\t" {return results}

        py eval [format $getModuleInfoFmt $moduleName]
        foreach moduleItem [py call pymanInspect] {
            lassign $moduleItem name type
            switch -glob -- $type {
                {<class *} {
                    ##
                    ## Define the class
                    ##
                    set className [format {%s::%s} [string trimright $namespaceName] $name]
                    ::oo::class create $className {
                        superclass ::pyman

                        self method init {attributes} {
                            variable Attributes

                            dict set Attributes list {}
                            dict set Attributes constants {}
                            dict set Attributes properties {}
                            foreach item $attributes {
                                lassign $item name value

                                if {[string index $name 0] eq "_"} {
                                    ##
                                    ## Skip private and magic attributes
                                    ##
                                    continue
                                }

                                dict lappend Attributes list $name
                                if {[string match {<property object at *} $value]} {
                                    ##
                                    ## Property that we can read/set
                                    ##
                                    dict set Attributes type $name property
                                    dict lappend Attributes properties $name
                                } elseif {[string match {<*} $value]} {
                                    ##
                                    ## Subobject Constant
                                    ##
                                    dict set Attributes type $name constant
                                    dict set Attributes constants $name type subobject
                                } elseif {[string equal -length 4 $value {set([*} ]} {
                                    ##
                                    ## Enumeration Constant
                                    ##
                                    dict set Attributes type $name constant
                                    dict set Attributes constants $name type enum
                                    set enumList {}
                                    foreach item [split [string range $value 5 end-2] ","] {
                                        lappend enumList [string range [string trim $item] 2 end-1]
                                    }
                                    dict set Attributes constants $name value $enumList
                                } elseif {[string match {{*}} $value]} {
                                    ##
                                    ## Symbol Value Enumeration Constants
                                    ##
                                    dict set Attributes type $name constant
                                    dict set Attributes constants $name type symbolics
                                    foreach item [split [string range $value 1 end-1] ","] {
                                        lassign [split $item ":"] symbol itemValue
                                        set symbol [string range [string trim $symbol] 2 end-1]
                                        set itemValue [string trim $itemValue]
                                        dict set Attributes constants $name value $symbol $itemValue
                                        dict set Attributes constants $name symbol $value $symbol
                                    }
                                } else {
                                    ##
                                    ## Simple Constant Value
                                    ##
                                    dict set Attributes type $name constant
                                    dict set Attributes constants $name type simple
                                    dict set Attributes constants $name value $value
                                }
                            }
                            dict set Attributes list [lsort -dictionary [dict get $Attributes list]]

                            ::oo::objdefine [self] deletemethod init
                        }

                        method _attributes {{pattern {*}}} {
                            classvariable Attributes

                            return [lsort [dict keys [dict get $Attributes type] $pattern]]
                        }

                        method _getAttribute {attribute} {
                            classvariable Attributes
                            variable pythonObject

                            if {![dict exists $Attributes type $attribute]} {
                                throw [list PYMAN UNKATTR $attribute] "unknown attribute '$attribute'"
                            }
                            switch -exact [dict get $Attributes type $attribute] {
                                constant {
                                    if {[dict get $Attributes constants $attribute type] eq "subobject"} {
                                        set pymanSetAttrFmt {def pymanGetAttr():}
                                        append pymanSetAttrFmt "\n\t" {return %1$s.%2$s}
                                        py eval [format $pymanSetAttrFmt $pythonObject $attribute]
                                        set result [py call pymanGetAttr]
                                    } else {
                                        set result [dict get $Attributes constants $attribute value]
                                    }
                                }
                                property {
                                    set pymanSetAttrFmt {def pymanGetAttr():}
                                    append pymanSetAttrFmt "\n\t" {return %1$s.%2$s}
                                    py eval [format $pymanSetAttrFmt $pythonObject $attribute]
                                    set result [py call pymanGetAttr]
                                }
                            }
                            return $result
                        }

                        method _setAttribute {attribute value} {
                            classvariable Attributes
                            variable pythonObject

                            if {![dict exists $Attributes type $attribute]} {
                                throw [list PYMAN UNKATTR $attribute] "unknown attribute '$attribute'"
                            }
                            switch -exact [dict get $Attributes type $attribute] {
                                constant {
                                    throw [list PYMAN SETCONST $attribute] "can not set constant attribute '$attribute'"
                                }
                                property {
                                    set pymanSetAttrFmt {def pymanSetAttr(newValue):}
                                    append pymanSetAttrFmt "\n\t" {%1$s.%2$s = newValue}
                                    py eval [format $pymanSetAttrFmt $pythonObject $attribute]
                                    py call pymanSetAttr $value
                                }
                            }
                        }
                    }

                    ##
                    ## Get attributes and their defaults
                    ##
                    py eval [format $pymanGetAttributesFmt $moduleName $name]
                    set attrList [py call pymanGetAttributes]
                    $className init $attrList

                    ##
                    ## Remove unwanted class methods
                    ##
                    #::oo::define $className deletemethod addPath
                    #::oo::define $className deletemethod import
                    ::oo::define $className export _attributes _getAttribute _setAttribute

                    ##
                    ## Introspect to see what methods exists and add them in
                    ##
                    py eval "pyman.getMethods = lambda : inspect.getmembers($moduleName.$name)"
                    foreach classItem [py call pyman.getMethods] {
                        lassign $classItem methodName methodInfo
                        if {[string match {<unbound method *} $methodInfo] && [string index $methodName 0] ne "_"} {
                            ::oo::define $className method $methodName args [format $methodBodyFmt $methodName]
                        }
                    }
                }
                {<function *} {
                    ##
                    ## Define the wrapper procedure
                    ##                            
                    if {[string index $name 0] eq "_"} {
                        ##
                        ## Skip private and magic methods
                        ##
                        continue
                    }
                    {*}[format {proc %1$s::%3$s {args} {py call %2$s.%3$s {*}$args}} $namespaceName $moduleName $name]
                }
            }
        }
    }

    classmethod addPath {pathName} {
        py eval "sys.path.append('$pathName')"
    }
    
}

py import inspect
py import sys
py import os
py import types
py eval {from types import ModuleType}
py eval {pyman = ModuleType('pyman')}
